;;; -*- mode:scheme; coding:utf-8; -*-
;;;
;;; text/json/compare.scm - JSON compare
;;;
;;;   Copyright (c) 2018  Takashi Kato  <ktakashi@ymail.com>
;;;
;;;   Redistribution and use in source and binary forms, with or without
;;;   modification, are permitted provided that the following conditions
;;;   are met:
;;;
;;;   1. Redistributions of source code must retain the above copyright
;;;      notice, this list of conditions and the following disclaimer.
;;;
;;;   2. Redistributions in binary form must reproduce the above copyright
;;;      notice, this list of conditions and the following disclaimer in the
;;;      documentation and/or other materials provided with the distribution.
;;;
;;;   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
;;;   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
;;;   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
;;;   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
;;;   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
;;;   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
;;;   TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
;;;   PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
;;;   LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
;;;   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
;;;   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
;;;

;; JSON object requires unordered comparison
#!nounbound
(library (text json compare)
    (export json=?)
    (import (rnrs)
	    (text json parse) ;; for *json-map-type*
	    (srfi :133 vectors))


(define (json=? a b)
  (case (*json-map-type*)
    ((vector) (vector-json=? a b))
    ((alist)  (alist-json=? a b))))

(define (vector-json=? a b)
  (define (entry=? a b)
    (and (vector-json=? (car a) (car b))
	 (vector-json=? (cdr a) (cdr b))))
  (define (key-compare a b) (string<? (car a) (car b)))
  (cond ((and (string? a) (string? b)) (string=? a b))
	((and (number? a) (number? b)) (= a b))
	((and (vector? a) (vector? b) (= (vector-length a) (vector-length b)))
	 (vector-every entry=?
		       (vector-sort key-compare a)
		       (vector-sort key-compare b)))
	((and (list? a) (list? b)) (for-all vector-json=? a b))
	(else (eq? a b))))

(define (alist-json=? a b)
  (define (entry=? a b)
    (and (alist-json=? (car a) (car b))
	 (alist-json=? (cdr a) (cdr b))))
  (define (key-compare a b) (string<? (car a) (car b)))
  (cond ((and (string? a) (string? b)) (string=? a b))
	((and (number? a) (number? b)) (= a b))
	((and (vector? a) (vector? b)) (vector-every alist-json=? a b))
	((and (list? a) (list? b) (= (length a) (length b)))
	 (for-all entry=? (list-sort key-compare a) (list-sort key-compare  b)))
	(else (eq? a b))))
)
